from __future__ import annotations  # Postponed eval of annotations. Fixed in 3.10
from typing import Optional

from .typedefs import IRectangle, ISelectionModel, IVec2, _translate_dict
from .scenenode import SceneNode
from .factories import (
    selection_factory,
    scene_factory,
    sceneitemfolder_factory,
    sceneitem_factory,
    scenenode_factory,
)

""" SelectionService and Selection share a lot in common.
    A mixin to support both. """


class SelectionBase:
    async def scene_id(self) -> str:
        response = await self._connection.command("sceneId", self._prepared_params())
        return response

    async def add(self, ids: list[str]):  # -> Selection
        response = await self._connection.command("add", self._prepared_params([ids]))
        return selection_factory(self._connection, response)

    async def center_on_screen(self) -> None:
        response = await self._connection.command(
            "centerOnScreen", self._prepared_params([])
        )
        self._check_empty(response)

    async def clone(self):  # -> Selection:
        response = await self._connection.command("clone", self._prepared_params())
        return selection_factory(self._connection, response)

    async def copy_to(
        self,
        scene_id: str,
        folder_id: Optional[str] = None,
        duplicate_sources: Optional[bool] = None,
    ) -> None:
        response = await self._connection.command(
            "copyTo", self._prepared_params([scene_id, folder_id, duplicate_sources])
        )
        self._check_empty(response)

    async def deselect(self, ids: list[str]):  # -> Selection:
        response = await self._connection.command(
            "deselect", self._prepared_params([ids])
        )
        return selection_factory(self._connection, response)

    async def fit_to_screen(self) -> None:
        response = await self._connection.command(
            "fitToScreen", self._prepared_params([])
        )
        self._check_empty(response)

    async def flip_x(self) -> None:
        response = await self._connection.command("flipX", self._prepared_params([]))
        self._check_empty(response)

    async def flip_y(self) -> None:
        response = await self._connection.command("flipY", self._prepared_params([]))
        self._check_empty(response)

    async def get_bounding_rect(self) -> IRectangle:
        response = await self._connection.command(
            "getBoundingRect", self._prepared_params([])
        )
        if response:
            return IRectangle(
                x=response["x"],
                y=response["y"],
                width=response["width"],
                height=response["height"],
            )
        else:
            return None

    async def get_folders(self) -> list[SceneItemFolder]:
        response = await self._connection.command(
            "getFolders", self._prepared_params([])
        )
        return [
            sceneitemfolder_factory(self._connection, subitem) for subitem in response
        ]

    async def get_ids(self) -> list[str]:
        response = await self._connection.command("getIds", self._prepared_params([]))
        return response

    async def get_inverted(self) -> list[SceneNode]:
        response = await self._connection.command(
            "getInverted", self._prepared_params([])
        )
        return [scenenode_factory(self._connection, subitem) for subitem in response]

    async def get_inverted_ids(self) -> list[str]:
        response = await self._connection.command(
            "getInvertedIds", self._prepared_params([])
        )
        return response

    async def get_items(self) -> list[SceneItem]:
        response = await self._connection.command("getItems", self._prepared_params([]))
        return [sceneitem_factory(self._connection, subitem) for subitem in response]

    async def get_last_selected(self) -> SceneNode:
        response = await self._connection.command(
            "getLastSelected", self._prtepared_params()
        )
        return scenenode_factory(self._connection, response)

    async def get_last_selected_id(self) -> bool:
        response = await self._connection.command(
            "getLastSelectedId", self._prepared_params()
        )
        return response

    async def get_model(self) -> ISelectionModel:
        response = await self._connection.command("getModel", self._prepared_params())
        return ISelectionModel(
            last_selected_id=response.get("lastSelectedId", []),
            selected_ids=response["selectedIds"],
        )

    async def get_root_nodes(self) -> list[SceneNode]:
        response = await self._connection.command(
            "getRootNodes", self._prepared_params()
        )
        return [scenenode_factory(self._connection, subitem) for subitem in response]

    async def get_scene(self) -> Scene:
        response = await self._connection.command("getScene", self._prepared_params())
        return scene_factory(self._connection, response)

    async def get_size(self) -> bool:
        response = await self._connection.command("getSize", self._prepared_params())
        return response

    async def get_sources(self) -> list[Sources]:
        response = await self._connection.command("getSources", self._prepared_params())
        return [source_factory(self._connection, subitem) for subitem in response]

    async def get_visual_items(self) -> list[SceneItems]:
        response = await self._connection.command(
            "getVisualItems", self._prepared_params()
        )
        return [sceneitem_factory(self._connection, subitem) for subitem in response]

    async def invert(self) -> Selection:
        response = await self._connection.command("invert", self._prepared_params())
        return selection_factory(self._connection, response)

    async def is_scene_folder(self) -> bool:
        response = await self._connection.command(
            "isSceneFolder", self._prepared_params()
        )
        return response

    async def is_scene_item(self) -> bool:
        response = await self._connection.command(
            "isSceneItem", self._prepared_params()
        )
        return response

    async def is_selected(self, node_id: str) -> bool:
        response = await self._connection.command(
            "isSelected", self._prepared_params([node_id])
        )
        return response

    async def move_to(self, scene_id: str, folder_id: Optional[str] = None) -> None:
        response = await self._connection.command(
            "moveTo", self._prepared_params([scene_id, folder_id])
        )
        self._check_empty(response)

    async def place_after(self, scene_node_id: str) -> None:
        response = await self._connection.command(
            "placeAfter", self._prepared_params([scene_node_id])
        )
        self._check_empty(response)

    async def place_before(self, scene_node_id: str) -> None:
        response = await self._connection.command(
            "placeBefore", self._prepared_params([scene_node_id])
        )
        self._check_empty(response)

    async def remove(self) -> None:
        response = await self._connection.command("remove", self._prepared_params([]))
        self._check_empty(response)

    async def reset(self) -> Selection:
        response = await self._connection.command("reset", self._prepared_params([]))
        return selection_factory(self._connection, response)

    async def reset_transform(self) -> None:
        response = await self._connection.command(
            "resetTransform", self._prepared_params([])
        )
        self._check_empty(response)

    async def rotate(self, deg: int) -> None:
        response = await self._connection.command(
            "rotate", self._prepared_params([deg])
        )
        self._check_empty(response)

    async def scale(self, scale: IVec2, origin: Optional[IVec2] = None) -> None:
        response = await self._connection.command(
            "scale", self._prepared_params([scale, origin] if origin else [scale])
        )
        self._check_empty(response)

    async def scale_with_offset(self, scale: IVec2, offset: IVec2) -> None:
        response = await self._connection.command(
            "scaleWithOffset", self._prepared_params([scale, offset])
        )
        self._check_empty(response)

    async def select(self, ids: list[ids]) -> Selection:
        response = await self._connection.command(
            "select", self._prepared_params([ids])
        )
        return selection_factory(self._connection, response)

    async def select_all(self) -> None:
        response = await self._connection.command(
            "selectAll", self._prepared_params([])
        )
        return ISelectionModel(
            last_selected_id=response.get("lastSelectedId", []),
            selected_ids=response["selectedIds"],
        )

    async def set_content_crop(self) -> None:
        response = await self._connection.command(
            "setContentCrop", self._prepared_params([])
        )
        self._check_empty(response)

    async def set_parent(self, folder_id: str) -> None:
        response = await self._connection.command(
            "setParent", self._prepared_params([folder_id])
        )
        self._check_empty(response)

    async def set_recording_visible(self, recording_visible: bool) -> None:
        response = await self._connection.command(
            "setRecordingVisible", self._prepared_params([recording_visible])
        )
        self._check_empty(response)

    async def set_settings(self, settings: dict) -> None:
        # Values in the settings:
        #   transform, visible, locked, stream_visible, recording_visible
        response = await self._connection.command(
            "setSettings", self._prepared_params([settings])
        )
        self._check_empty(response)

    async def set_stream_visible(self, stream_visible: bool) -> None:
        response = await self._connection.command(
            "setStreamVisible", self._prepared_params([stream_visible])
        )
        self._check_empty(response)

    async def set_transform(self, transform: dict) -> None:
        # The definition of IPartialTransform is a dictionary that may optionally
        # contain:
        #    position (2-tuple)
        #    scale (2-tuple)
        #    crop (a dictionary that may contain top, bottom, left, right mapping to
        #           ints)
        #    rotation
        translated_dict = _translate_dict(transform)
        response = await self._connection.command(
            "setTransform", self._prepared_params([translated_dict])
        )
        self._check_empty(response)

    async def set_visibility(self, visible: boolean) -> None:
        response = await self._connection.command(
            "setVisibility", self._prepared_params([visible])
        )
        self._check_empty(response)

    async def stretch_to_screen(self) -> None:
        response = await self._connection.command(
            "stretchToScreen", self._prepared_params([])
        )
        self._check_empty(response)



from connection import AuthenticationFailure, ProtocolError, SlobsConnection
from pubsubhub import SubscriptionPreferences
from slobs.audioservice import AudioService
from slobs.notificationsservice import NotificationsService
from slobs.performanceservice import PerformanceService
from slobs.scenecollectionservice import SceneCollectionsService
from slobs.selectionservice import SelectionService
from slobs.sourcesservice import SourcesService
from slobs.streamingservice import StreamingService
from slobs.transitionsservice import TransitionsService
from slobs.typedefs import NotificationSubType, NotificationType



"""
    The SLOBS API has concepts of
        * "Service"
            - top level entities can be instantiated at any time. They
              support methods and subscriptions.
        * "Class"
            - entities that are only instantiated by being returned from the
              server. They support methods and have some data attached.
        * "Interfaces"
            - entities that contain data but support no methods.
        * "Types"
            - restrictions on parameter values, like enumerated types.

    The Python API represents these internally as:
        * Service -> SlobsService
            - defined in this file.
        * Class -> SlobsClass
            - defined in this file.
        * Interface -> namedtuple
            - defined with the associated SlobsClasses
        * Types -> Enum types (and bool, str, int, float)
            - defined with the associated SlobsClasses
"""

from .connection import ProtocolError
from .pubsubhub import SubscriptionPreferences

"""
    Design note: Considered the use of CompactMode, which limits the number
    of fields returned when a new SlobsClass is introduced. 

    Pro: 
        * less processing per message
        * it avoids the issue of how to cache this data which may become
          obsolete without notice.

    Cons:
        * more round trips to gather basic information.
        * there doesn't appear to be a way to populate name fields otherwise.

    Compromise:
        * get all the fields, but only use the ones that appear as properties
          on the SlobsClasses.
"""


class SlobsBase:
    def __init__(self, connection, resource_id):
        self._connection = connection
        self._resource_id = resource_id

    def __str__(self):
        return f"{self.__class__.__name__}({self._resource_id})"

    # Helper functions.

    def _prepared_params(self, args=None):
        if args is None:
            args = []
        else:
            assert isinstance(args, list)
        return dict(resource=self._resource_id, args=args)

    @staticmethod
    def _check_empty(response):
        if response:
            raise ProtocolError("Expected void response: %s", response)


class SlobsService(SlobsBase):
    def __init__(self, connection):
        # This takes advantage of the fact that the Python class names
        # match the Javascript Resource names.
        super().__init__(connection, resource_id=self.__class__.__name__)


class SlobsClass(SlobsBase):
    """Base class for proxies to "Classes" in the data model

    Attributes are read-only, and are at risk of being out-of-date
    if there are changes to the server or to other instances referring
    to the same API entity.
    """

    """ 
        Design note: Ids.

        It doesn't appear to be well documented, but each SlobsClass has a
            source_id - a unique string, and a
            resource_id - which appears to be of the pp class_name["source_id"].

        Some SLobsClass also have an (unrelated?) field called "id".

        Each service just has a resource_id which is just the class name.

        It seems when referring to yourself, you have to provide the resource_id.
        When referencing associated SlobsClasses, you have to provide an "id" which 
        appears to actually be referring to the source_id.

    """

    def __init__(self, connection, resource_id, source_id):
        super().__init__(connection, resource_id)
        self._source_id = source_id

    @property
    def source_id(self):
        return self._source_id

    @property
    def resource_id(self):
        return self._resource_id

    def __str__(self):
        # Override to use source_id instead.
        return f"{self.__class__.__name__}({self._source_id})"


class _EventSubscription:
    def __init__(self, connection, resource_id, callback_coroutine):
        self._connection = connection
        self._resource_id = resource_id
        self._callback_coroutine = callback_coroutine

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.unsubscribe()

    async def unsubscribe(self):
        await self._connection.unsubscribe(self._resource_id, self._callback_coroutine)


class Event:
    """An event that can be subscribed to"""

    def __init__(self, connection, method: str, service: str):
        self._connection = connection
        self._service = service
        self._method = method

    async def subscribe(
        self, callback_coroutine, subscription_prefs=SubscriptionPreferences()
    ) -> _EventSubscription:
        subscription_resource_id = await self._connection.subscribe(
            method=self._method,
            params=dict(resource=self._service, args=[]),
            callback_coroutine=callback_coroutine,
            subscription_preferences=subscription_prefs,
        )

        return _EventSubscription(
            self._connection,
            subscription_resource_id,
            callback_coroutine,
        )



from ..apibase import SlobsService, Event
from .typedefs import ITransitionsServiceState


class TransitionsService(SlobsService):
    def __init__(self, connection):
        super().__init__(connection)

        self.studio_mode_changed = Event(
            connection, "studioModeChanged", self.__class__.__name__
        )

    async def disable_studio_mode(self):
        # Warning: Intermittently fails with ProtocolError
        response = await self._connection.command(
            "disableStudioMode", self._prepared_params()
        )
        self._check_empty(response)

    async def enable_studio_mode(self):
        # Warning: Intermittently fails with ProtocolError
        response = await self._connection.command(
            "enableStudioMode", self._prepared_params()
        )
        self._check_empty(response)

    async def execute_studio_mode_transition(self):
        response = await self._connection.command(
            "executeStudioModeTransition", self._prepared_params()
        )
        self._check_empty(response)

    async def get_model(self) -> ITransitionsServiceState:
        response = await self._connection.command("getModel", self._prepared_params())
        return ITransitionsServiceState(studio_mode=response["studioMode"])



from pathlib import Path
from typing import Optional
import configparser
from dataclasses import dataclass

DEFAULT_DOMAIN = "localhost"
DEFAULT_PORT = 59650


@dataclass
class ConnectionConfig:
    """Information required to connect to SL OBS."""

    token: str
    domain: str = DEFAULT_DOMAIN
    port: int = DEFAULT_PORT


def config_from_ini() -> Optional[ConnectionConfig]:
    """Read ConnectionConfig from ini files.
    Looks in:
        - <current dir>/pyslobs.ini
        - <current dir>/.pyslobs
        - <home dir>/pyslobs.ini
        - <home dir>/.pyslobs
    May return None if no ini file is found.
    """
    ini_file_paths = [
        Path.cwd() / "pyslobs.ini",
        Path.cwd() / ".pyslobs",
        Path.home() / "pyslobs.ini",
        Path.home() / ".pyslobs",
    ]
    parser = configparser.ConfigParser()
    if not parser.read(ini_file_paths):
        # No ini file
        return None
    assert (
        "connection" in parser.sections()
    ), "Ini file contains no 'connection' section"
    assert "token" in parser["connection"], "Ini file contains no 'token' value"
    token = parser["connection"]["token"]
    domain = parser["connection"].get("domain", DEFAULT_DOMAIN)
    try:
        port = int(parser["connection"].get("port", DEFAULT_PORT))
    except ValueError:
        assert False, "Port in ini file is not an integer"

    return ConnectionConfig(token, domain, port)


def config_from_ini_else_stdin() -> ConnectionConfig:
    """Read ConnectionConfig from ini files, but if they are absent,
    ask for API token on StdIn."""
    return config_from_ini() or ConnectionConfig(input("API Token: "))



from typing import Optional, Dict, List

from ..apibase import SlobsService, Event
from .source import Source
from .typedefs import TSourceType, IObsListOption, ISourceAddOptions
from .factories import source_factory


class SourcesService(SlobsService):
    def __init__(self, connection):
        super().__init__(connection)

        self.source_added = Event(connection, "sourceAdded", self.__class__.__name__)

        self.source_removed = Event(
            connection, "sourceRemoved", self.__class__.__name__
        )

        self.source_updated = Event(
            connection, "sourceUpdated", self.__class__.__name__
        )

    async def add_file(self, path: str) -> Source:
        response = await self._connection.command(
            "addFile", self._prepared_params([str(path)])
        )
        return source_factory(self._connection, response)

    async def create_source(
        self,
        name: str,
        type_: TSourceType,
        settings: Optional[Dict] = None,
        options: Optional[ISourceAddOptions] = None,
    ) -> Source:
        options_dict = (
            {}
            if not options
            else {"channel": options.channel, "isTemporary": options.is_temporary}
        )
        response = await self._connection.command(
            "createSource",
            self._prepared_params([name, type_, settings, options_dict]),
        )
        return source_factory(self._connection, response)

    async def get_available_sources_types_list(self) -> list[any]:
        response = await self._connection.command(
            "getAvailableSourcesTypesList", self._prepared_params([])
        )
        return [
            IObsListOption(value=subitem["value"], description=subitem["description"])
            for subitem in response
        ]

    async def get_source(self, source_id: str) -> Source:
        response = await self._connection.command(
            "getSource", self._prepared_params([source_id])
        )
        return source_factory(self._connection, response)

    async def get_sources(self) -> List[Source]:
        response = await self._connection.command(
            "getSources", self._prepared_params([])
        )
        return [source_factory(self._connection, subitem) for subitem in response]

    async def get_sources_by_name(self, name: str) -> List[Source]:
        response = await self._connection.command(
            "getSourcesByName", self._prepared_params([name])
        )
        return [source_factory(self._connection, subitem) for subitem in response]

    async def remove_source(self, id_: str) -> None:
        response = await self._connection.command(
            "removeSource", self._prepared_params([id_])
        )
        self._check_empty(response)

    async def show_add_source(self, source_type: TSourceType) -> None:
        """
        Causes the UI to open the Add Source dialog box, at the point where
        the source-type has been selected and a name is to be entered.
        """
        response = await self._connection.command(
            "showAddSource", self._prepared_params([source_type])
        )
        self._check_empty(response)

    async def show_showcase(self) -> None:
        """
        Causes the UI to open the Add Source dialog box, at the point where
        the source-type is to be selected.
        """
        response = await self._connection.command(
            "showShowcase", self._prepared_params([])
        )
        self._check_empty(response)

    async def show_source_properties(self, source_id: str) -> Source:
        """
        Causes the UI to open the settings dialog box on the selected source.
        """
        response = await self._connection.command(
            "showSourceProperties", self._prepared_params([source_id])
        )
        self._check_empty(response)



"""
    When a json dict is received in a reply, we create a Python instance
    equivalent.

    This code should not be used by the client, but it is shared between a
    number of SlobsService and SlobsClasses so it has been grouped together.
"""
from .typedefs import TSceneNodeType, ITransform, IVec2, ICrop

CLASSES = {}

# DependencyInjection to avoid circular dependencies.
def register(cls):
    CLASSES[cls.__name__] = cls


def audiosource_factory(conn, json_dict):
    return CLASSES["AudioSource"](
        conn,
        source_id=json_dict["sourceId"],
        resource_id=json_dict["resourceId"],
    )


def scenenode_factory(connection, json_dict):
    if json_dict["sceneNodeType"] == "item":
        return sceneitem_factory(connection, json_dict)
    else:
        assert json_dict["sceneNodeType"] == "folder"
        return sceneitemfolder_factory(connection, json_dict)


def sceneitem_factory(connection, json_dict):
    # We see an inconsistency in the API here.
    # Sometimes the position is a list [x, y] and sometimes a dict(x=_, y=_)
    if isinstance(json_dict["transform"]["position"], list):
        transform_position = (
            IVec2(
                x=json_dict["transform"]["position"][0],
                y=json_dict["transform"]["position"][1],
            ),
        )
    else:
        transform_position = IVec2(**json_dict["transform"]["position"])

    return CLASSES["SceneItem"](
        connection=connection,
        resource_id=json_dict["resourceId"],
        source_id=json_dict["sourceId"],
        id_=json_dict["id"],
        # Spec says nodeId should be present, but found it wasn't.
        node_id=json_dict.get("nodeId", None),
        parent_id=json_dict["parentId"],
        scene_id=json_dict["sceneId"],
        scene_node_type=TSceneNodeType(json_dict["sceneNodeType"]),
        name=json_dict["name"],
        locked=json_dict["locked"],
        recording_visible=json_dict["recordingVisible"],
        scene_item_id=json_dict["sceneItemId"],
        stream_visible=json_dict["streamVisible"],
        transform=ITransform(
            crop=ICrop(**json_dict["transform"]["crop"]),
            position=transform_position,
            rotation=json_dict["transform"]["rotation"],
            scale=IVec2(**json_dict["transform"]["scale"]),
        ),
        visible=json_dict["visible"],
    )


def sceneitemfolder_factory(connection, json_dict):
    # Found children_ids as part of the dict, but not the spec.
    # Might like to add in if useful to avoid a call.
    return CLASSES["SceneItemFolder"](
        connection=connection,
        # resourceId/sourceId oddly not part of the returned dict.
        # fall back to id? Doesn't seem to work.
        resource_id=json_dict.get("resourceId", json_dict["id"]),
        source_id=json_dict.get("sourceId"),
        id_=json_dict["id"],
        name=json_dict["name"],
        # Spec says nodeId should be present, but found it wasn't.
        node_id=json_dict.get("nodeId", None),
        parent_id=json_dict["parentId"],
        scene_id=json_dict["sceneId"],
        scene_node_type=TSceneNodeType(json_dict["sceneNodeType"]),
    )


def source_factory(connection, json_dict):
    return CLASSES["Source"](
        connection,
        # resource_ids are missing after a deletion.
        resource_id=json_dict.get("resourceId", None),
        source_id=json_dict.get("sourceId", None),
        async_=json_dict["async"],
        audio=json_dict["audio"],
        channel=json_dict.get("channel", None),
        configurable=json_dict.get("configurable", None),
        do_not_duplicate=json_dict["doNotDuplicate"],
        height=json_dict["height"],
        id_=json_dict["id"],
        muted=json_dict["muted"],
        name_=json_dict["name"],
        type_=json_dict["type"],
        video=json_dict["video"],
        width=json_dict["width"],
    )


def scene_factory(connection, json_dict):
    return CLASSES["Scene"](
        connection,
        # resource_ids are missing after a deletion.
        resource_id=json_dict.get("resourceId", None),
        source_id=json_dict["id"],
        # Names are sometimes missing. Fill in an empty name.
        name=json_dict.get("name", ""),
        id=json_dict["id"],
        nodes=[
            # Having trouble with these nodes not being gettable. Not sure why.
            scenenode_factory(connection, node)
            for node in json_dict["nodes"]
        ],
    )


def selection_factory(connection, json_dict):
    return CLASSES["Selection"](
        connection,
        # resource_ids are missing after a deletion.
        resource_id=json_dict.get("resourceId", None),
    )



from ..apibase import SlobsService
from .typedefs import IPerformanceState


class PerformanceService(SlobsService):
    async def get_model(self):
        response = await self._connection.command("getModel", self._prepared_params())
        return IPerformanceState(
            cpu=response["CPU"],
            bandwidth=response["numberDroppedFrames"],
            frame_rate=response["numberDroppedFrames"],
            number_dropped_frames=response["numberDroppedFrames"],
            percentage_dropped_frames=response["percentageDroppedFrames"],
        )



from ..apibase import SlobsBase
from .selectionbase import SelectionBase
from .factories import register

# Design note: In practice, Selections do not have real source_ids or ids, and have a
# resourceId of 'Selection[]'.
# Rather than calling them an instance of SlobsClass, falling back to SlobsBase.

{
    "_type": "HELPER",
    "resourceId": "Selection[]",
    "lastSelectedId": "d85a672a-41a2-49c5-8e7f-201cd8dbfbec",
    "selectedIds": [
        "ccefc122-e8e6-4982-b2e6-6dd4e5da70ef",
        "55a35a1a-a5b3-42bd-9df7-b6cd6bae020f",
        "d85a672a-41a2-49c5-8e7f-201cd8dbfbec",
    ],
}


class Selection(SlobsBase, SelectionBase):
    def __init__(self, connection, resource_id):
        SlobsBase.__init__(self, connection, resource_id=resource_id)
        SelectionBase.__init__(self)


register(Selection)



from datetime import datetime
from typing import Dict

from ..apibase import SlobsService
from .typedefs import (
    INotificationModel,
    NotificationType,
    INotificationSettings,
    _translate_dict,
    INOTIFICATIONOPTIONS_FIELDS,
    INOTIFICATIONOPTIONS_MANDATORY_FIELDS,
    INOTIFICATIONSETTINGS_FIELDS,
    NotificationSubType,
)


class NotificationsService(SlobsService):
    @staticmethod
    def _inotification_model_factory(json_dict):
        return INotificationModel(
            action=json_dict.get("action", None),
            code=json_dict.get("code", None),
            data=json_dict.get("data", None),
            date=datetime.fromtimestamp(json_dict["date"] / 1000),
            id=json_dict["id"],
            lifetime=json_dict["lifeTime"],
            message=json_dict["message"],
            play_sound=json_dict["playSound"],
            show_time=json_dict["showTime"],
            subtype=NotificationSubType(json_dict["subType"]),
            type=NotificationType(json_dict["type"]),
            unread=json_dict["unread"],
        )

    async def apply_action(self, notification_id):
        response = await self._connection.command(
            "applyAction", self._prepared_params([notification_id])
        )
        self._check_empty(response)

    async def get_all(self, type_: NotificationType):
        response = await self._connection.command(
            "getAll", self._prepared_params([type_.value])
        )
        return [self._inotification_model_factory(subitem) for subitem in response]

    async def get_notification(self, id_):
        response = await self._connection.command(
            "getNotification", self._prepared_params([id_])
        )
        return self._inotification_model_factory(response)

    async def get_read(self, type_: NotificationType):
        response = await self._connection.command(
            "getRead", self._prepared_params([type_.value])
        )
        return [self._inotification_model_factory(subitem) for subitem in response]

    async def get_settings(self) -> INotificationSettings:
        response = await self._connection.command(
            "getSettings", self._prepared_params()
        )
        return INotificationSettings(
            enabled=response["enabled"], play_sound=response["playSound"]
        )

    async def get_unread(self, type_: NotificationType):
        response = await self._connection.command(
            "getUnread", self._prepared_params([type_.value])
        )
        return [self._inotification_model_factory(subitem) for subitem in response]

    async def mark_all_as_read(self):
        response = await self._connection.command(
            "markAllAsRead", self._prepared_params()
        )
        self._check_empty(response)

    async def mark_as_read(self, id_):
        response = await self._connection.command(
            "markAsRead", self._prepared_params([id_])
        )
        self._check_empty(response)

    async def push(self, notify_info: Dict):
        translated_dict = _translate_dict(notify_info)
        # notify_info may only contain fields in the INotificationOptions list
        assert set(translated_dict.keys()).issubset(INOTIFICATIONOPTIONS_FIELDS)
        # notify_info must contain mandatory fields
        assert INOTIFICATIONOPTIONS_MANDATORY_FIELDS.issubset(
            set(translated_dict.keys())
        )

        response = await self._connection.command(
            "push", self._prepared_params([translated_dict])
        )
        return self._inotification_model_factory(response)

    async def restore_default_settings(self):
        response = await self._connection.command(
            "restoreDefaultSettings", self._prepared_params()
        )
        self._check_empty(response)

    async def set_settings(self, patch: Dict):
        # patch should be a dictionary containing any subset of the available
        # fields as keys.
        _translated_dict = _translate_dict(patch)
        assert set(_translated_dict.keys()).issubset(set(INOTIFICATIONSETTINGS_FIELDS))
        response = await self._connection.command(
            "setSettings", self._prepared_params([_translated_dict])
        )
        self._check_empty(response)

    async def show_notifications(self):
        response = await self._connection.command(
            "showNotifications", self._prepared_params()
        )
        self._check_empty(response)



from __future__ import annotations

from .factories import (
    scenenode_factory,
    sceneitem_factory,
    sceneitemfolder_factory,
    scene_factory,
    selection_factory,
    source_factory,
    register,
)
from ..apibase import SlobsClass
from .typedefs import ITransform, TSceneNodeType, ISceneNodeModel, IVec2


class SceneNode(SlobsClass):
    def __init__(
        self,
        connection,
        resource_id: str,
        source_id: str,
        id_: str,
        node_id: str,
        parent_id: str,
        scene_id: str,
        scene_node_type: TSceneNodeType,
    ):
        super().__init__(connection, resource_id=resource_id, source_id=source_id)

        self._id = id_
        self._node_id = node_id
        self._parent_id = parent_id
        self._scene_id = scene_id
        self._scene_node_type = scene_node_type

    def __str__(self):
        return f"{self.__class__.__name__}(resource_id={self._resource_id}, id_={self._id})"

    @property
    def id_(self):
        return self._id

    @property
    def node_id(self):
        return self._node_id

    @property
    def parent_id(self):
        return self._parent_id

    @property
    def scene_id(self):
        return self._scene_id

    @property
    def scene_node_type(self):
        return self._scene_node_type

    async def add_to_selection(self) -> None:
        response = await self._connection.command(
            "addToSelection", self._prepared_params()
        )
        self._check_empty(response)

    async def deselect(self) -> None:
        response = await self._connection.command("deselect", self._prepared_params())
        self._check_empty(response)

    async def detach_parent(self) -> None:
        response = await self._connection.command(
            "detachParent", self._prepared_params()
        )
        self._check_empty(response)

    async def get_item_index(self) -> int:
        response = await self._connection.command(
            "getItemIndex", self._prepared_params()
        )
        return response

    async def get_model(self):  # -> ISceneNodeModel:
        response = await self._connection.command("getModel", self._prepared_params())
        return scenenode_factory(response)

    async def get_next_item(self):  # -> ISceneItem:
        response = await self._connection.command(
            "getNextItem", self._prepared_params()
        )
        return sceneitem_factory(response)

    async def get_next_node(self):  # -> ISceneNodeModel:
        response = await self._connection.command(
            "getNextNode", self._prepared_params()
        )
        return scenenode_factory(response)

    async def get_node_index(self) -> int:
        response = await self._connection.command(
            "getNodeIndex", self._prepared_params()
        )
        return response

    async def get_parent(self) -> ISceneNodeModel | None:
        response = await self._connection.command("getParent", self._prepared_params())
        if response:
            return sceneitemfolder_factory(self._connection, response)
        else:
            # Sometimes returns nothing, which isn't in the spec.
            return None

    async def get_path(self) -> str:
        response = await self._connection.command("getPath", self._prepared_params())
        return response

    async def get_prev_item(self):  # -> ISceneItem:
        response = await self._connection.command(
            "getPrevItem", self._prepared_params()
        )
        return sceneitem_factory(response)

    async def get_prev_node(self) -> ISceneNodeModel:
        response = await self._connection.command(
            "getPrevNode", self._prepared_params()
        )
        return scenenode_factory(response)

    async def get_scene(self):
        response = await self._connection.command("getScene", self._prepared_params())
        return scene_factory(self._connection, response)

    async def has_parent(self) -> bool:
        response = await self._connection.command("hasParent", self._prepared_params())
        return response

    async def is_folder(self) -> bool:
        response = await self._connection.command("isFolder", self._prepared_params())
        return response

    async def is_item(self):
        response = await self._connection.command("isItem", self._prepared_params())
        return response

    async def is_selected(self) -> bool:
        response = await self._connection.command("isSelected", self._prepared_params())
        return response

    async def place_after(self, node_id) -> None:
        response = await self._connection.command(
            "placeAfter", self._prepared_params([node_id])
        )
        self._check_empty(response)

    async def place_before(self, node_id) -> None:
        response = await self._connection.command(
            "placeBefore", self._prepared_params([node_id])
        )
        self._check_empty(response)

    async def remove(self) -> None:
        response = await self._connection.command("remove", self._prepared_params())
        self._check_empty(response)

    async def select(self) -> None:
        response = await self._connection.command("select", self._prepared_params())
        self._check_empty(response)

    async def set_parent(self, parent_id: str) -> None:
        response = await self._connection.command(
            "setParent", self._prepared_params([parent_id])
        )
        self._check_empty(response)


class SceneItemFolder(SceneNode):
    def __init__(
        self,
        connection,
        resource_id: str,
        source_id: str,
        id_: str,
        node_id: str,
        parent_id: str,
        scene_id: str,
        scene_node_type: TSceneNodeType,
        name: str,
    ):
        super().__init__(
            connection,
            resource_id,
            source_id,
            id_,
            node_id,
            parent_id,
            scene_id,
            scene_node_type,
        )
        self._name = name

    @property
    def scene_node_type(self):
        return self._scene_node_type

    @property
    def name(self):
        return self._name

    async def add(self, scene_node_id: str):
        response = await self._connection.command(
            "add", self._prepared_params([scene_node_id])
        )
        self._check_empty(response)

    async def get_folders(self) -> list[SceneItemFolder]:
        response = await self._connection.command("getFolders", self._prepared_params())
        return [
            sceneitemfolder_factory(self._connection, json_dict)
            for json_dict in response
        ]

    async def get_items(self) -> list[SceneItem]:
        response = await self._connection.command("getItems", self._prepared_params())
        return [
            sceneitem_factory(self._connection, json_dict) for json_dict in response
        ]

    async def get_nested_nodes(self) -> list[SceneItem]:
        response = await self._connection.command(
            "getNestedNodes", self._prepared_params()
        )
        return [
            scenenode_factory(self._connection, json_dict) for json_dict in response
        ]

    async def get_nodes(self) -> list[SceneItem]:
        response = await self._connection.command("getNodes", self._prepared_params())
        return [
            scenenode_factory(self._connection, json_dict) for json_dict in response
        ]

    async def get_selection(self) -> Selection:
        response = await self._connection.command(
            "getSelection", self._prepared_params()
        )
        return selection_factory(self._connection, json_dict)

    async def set_name(self, new_name: str):
        response = await self._connection.command(
            "setName", self._prepared_params([new_name])
        )
        self._check_empty(response)
        self._name = new_name

    async def ungroup(self):
        response = await self._connection.command("ungroup", self._prepared_params())
        self._check_empty(response)


class SceneItem(SceneNode):
    def __init__(
        self,
        connection,
        resource_id: str,
        source_id: str,
        id_: str,
        node_id: str,
        parent_id: str,
        scene_id: str,
        scene_node_type: TSceneNodeType,
        name: str,
        locked: bool,
        recording_visible: bool,
        scene_item_id: str,
        stream_visible: bool,
        transform: ITransform,
        visible: bool,
    ):
        super().__init__(
            connection,
            resource_id,
            source_id,
            id_,
            node_id,
            parent_id,
            scene_id,
            scene_node_type,
        )

        self._locked = locked
        self._name = name
        self._recording_visible = recording_visible
        self._scene_item_id = scene_item_id
        self._stream_visible = stream_visible
        self._transform = transform
        self._visible = visible

    @property
    def locked(self):
        return self._locked

    @property
    def name(self):
        return self._name

    @property
    def recording_visible(self):
        return self._recording_visible

    @property
    def scene_item_id(self):
        return self._scene_item_id

    @property
    def stream_visible(self):
        return self._stream_visible

    @property
    def transform(self):
        return self._transform

    @property
    def visible(self):
        return self._visible

    async def center_on_screen(self):
        response = await self._connection.command(
            "centerOnScreen", self._prepared_params()
        )
        self._check_empty(response)

    async def fit_to_screen(self):
        response = await self._connection.command(
            "fitToScreen", self._prepared_params()
        )
        self._check_empty(response)

    async def flip_x(self):
        response = await self._connection.command("flipX", self._prepared_params())
        self._check_empty(response)

    async def flip_y(self):
        response = await self._connection.command("flipY", self._prepared_params())
        self._check_empty(response)

    async def get_source(self):
        response = await self._connection.command("getSource", self._prepared_params())
        return source_factory(self._connection, response)

    async def reset_transform(self):
        response = await self._connection.command(
            "resetTransform", self._prepared_params()
        )
        self._check_empty(response)

    async def rotate(self, deg):
        response = await self._connection.command(
            "rotate", self._prepared_params([deg])
        )
        self._check_empty(response)

    async def set_content_crop(self):
        response = await self._connection.command(
            "setContentCrop", self._prepared_params()
        )
        self._check_empty(response)

    async def set_scale(
        self, new_scale_model: IVec2, origin: Option[IVec2] = None
    ) -> None:
        scm_params = {
            "x": new_scale_model.x,
            "y": new_scale_model.y,
        }
        if not origin:
            params = [scm_params]
        else:
            origin_params = {
                "x": origin.x,
                "y": origin.y,
            }
            params = [scm_params, origin_params]

        response = await self._connection.command(
            "setScale", self._prepared_params(params)
        )
        self._check_empty(response)
        self._transform = ITransform(
            crop=self._transform.crop,
            position=origin if origin else self.transform.position,
            rotation=self._transform.rotation,
            scale=new_scale_model,
        )

    async def set_settings(self, settings: dict[Any]) -> None:
        params = {}
        if settings.get("locked", None) is not None:
            params["locked"] = settings["locked"]
            self._locked = settings["locked"]
        if settings.get("recording_visible", None) is not None:
            params["recordingVisible"] = settings["recording_visible"]
            self._recording_visible = settings["recording_visible"]
        if settings.get("stream_visible", None) is not None:
            params["streamVisible"] = settings["stream_visible"]
            self._stream_visible = settings["stream_visible"]
        if settings.get("transform", None) is not None:
            params["transform"] = self.marshall_transform(settings["transform"])
            self._transform = settings["transform"]
        if settings.get("visible", None) is not None:
            params["visible"] = settings["visible"]
            self._visible = settings["visible"]

        response = await self._connection.command(
            "setSettings",
            self._prepared_params([params]),
        )
        self._check_empty(response)

    async def set_transform(self, transform: ITransform):
        response = await self._connection.command(
            "setTransform",
            self._prepared_params([self.marshall_transform(transform)]),
        )
        self._check_empty(response)

    async def set_visibility(self, visible):
        response = await self._connection.command(
            "setVisibility", self._prepared_params([visible])
        )
        self._visible = visible
        self._check_empty(response)

    async def stretch_to_screen(self):
        response = await self._connection.command(
            "stretchToScreen", self._prepared_params()
        )
        self._check_empty(response)

    @staticmethod
    def marshall_transform(transform):
        return {
            "crop": None
            if not transform.crop
            else dict(
                bottom=transform.crop.bottom,
                top=transform.crop.top,
                left=transform.crop.left,
                right=transform.crop.right,
            ),
            "position": None
            if not transform.position
            else dict(x=transform.position.x, y=transform.position.y),
            "rotation": transform.rotation,
            "scale": None
            if not transform.scale
            else dict(x=transform.scale.x, y=transform.scale.y),
        }


register(SceneNode)
register(SceneItem)
register(SceneItemFolder)



from ..apibase import SlobsService
from .selectionbase import SelectionBase
from .selection import Selection  # For the side-effects


class SelectionService(SlobsService, SelectionBase):
    def __init__(self, connection):
        SlobsService.__init__(self, connection)
        SelectionBase.__init__(self)



from collections import namedtuple
from datetime import datetime
from enum import Enum

_FIELD_MAPS = {
    "life_time": "lifeTime",
    "lifetime": "lifeTime",
    "play_sound": "playSound",
    "recording_visible": "recordingVisible",
    "show_time": "showTime",
    "stream_visible": "streamVisible",
    "subtype": "subType",
    "sub_type": "subType",
    }

# Translate from Pythonic dictionary keys to Javascripty ones.
# Only required for dictionaries passed to API by client.
def _translate_dict(raw_dict):
    return dict([(_FIELD_MAPS.get(key, key), value) for key, value in raw_dict.items()])


class NotificationType(Enum):
    INFO = "INFO"
    SUCCESS = "SUCCESS"
    WARNING = "WARNING"


class NotificationSubType(Enum):
    DEFAULT = "DEFAULT"
    DISCONNECTED = "DISCONNECTED"
    DROPPED = "DROPPED"
    LAGGED = "LAGGED"
    SKIPPED = "SKIPPED"


INotificationModel = namedtuple(
    "INotificationModel",
    "action code data date id lifetime message play_sound "
    "show_time  subtype  type  unread",
)
INOTIFICATIONOPTIONS_MANDATORY_FIELDS = {"message"}
INOTIFICATIONOPTIONS_OPTIONAL_FIELDS = {
    "action",
    "code",
    "data",
    "lifeTime",
    "playSound",
    "showTime",
    "subType",
    "type",
    "unread",
}
INOTIFICATIONOPTIONS_FIELDS = (
    INOTIFICATIONOPTIONS_MANDATORY_FIELDS | INOTIFICATIONOPTIONS_OPTIONAL_FIELDS
)
INOTIFICATIONSETTINGS_FIELDS = ["enabled", "playSound"]
INotificationSettings = namedtuple("INotificationSettings", ["enabled", "play_sound"])
IPerformanceState = namedtuple(
    "IPerformanceState",
    "cpu bandwidth frame_rate number_dropped_frames percentage_dropped_frames",
)
ISceneModel = namedtuple("ISceneModel", "id name nodes")
IStreamingState = namedtuple(
    "IStreamingState",
    "recording_status recording_status_time replay_buffer_status "
    "replay_buffer_status_time streaming_status streaming_status_time",
)


def _convert_time(datestr):
    return datetime.strptime(datestr[:-1] + "000Z", "%Y-%m-%dT%H:%M:%S.%f%z")


ITransitionsServiceState = namedtuple("ITransitionsServiceState", "studio_mode")


class MonitoringType(Enum):
    NONE = 0
    MONITORING_ONLY = 1
    MONITORING_AND_OUTPUT = 2


IFader = namedtuple("IFader", "db deflection mul")
IAudioSourceModel = namedtuple(
    "IAudioSourceModel",
    "audio_mixers fader force_mono mixer_hidden monitoring_type muted name "
    "source_id sync_offset",
)


IRectangle = namedtuple("IRectangle", "x y width height")


class TSceneNodeType(Enum):
    FOLDER = "folder"
    ITEM = "item"


ISceneNodeModel = namedtuple(
    "ISceneNodeModel", "children_ids id parent_id scene_id scene_node_type"
)

ISceneNodeAddOptions = namedtuple("ISceneNodeAddOptions", "id source_add_options")

ISourceAddOptions = namedtuple("ISourceAddOptions", "channel is_temporary")

ITransform = namedtuple("ITransform", "crop position rotation scale")

IObsListOption = namedtuple("IObsListOptions", "value description")

# These is unclearly defined.
TObsFormData = str

# This is one of the source types, like browser_source.
# The master list is returned by SourcesServices.getAvailableSourcesTypesList()
TSourceType = str

ISourceModel = namedtuple(
    "ISourceModel",
    [
        "async_",
        "audio",
        "channel",
        "do_not_duplicate",
        "height",
        "id",
        "muted",
        "name",
        "source_id",
        "type_",
        "video",
        "width",
    ],
)

ISceneCollectionsManifestEntry = namedtuple("ISceneCollectionsManifestEntry", "id name")

ISceneCollectionSchema = namedtuple("ISceneCollectionSchema", "id name scenes sources")

ISceneCollectionCreateOptions = namedtuple("ISceneCollectionCreateOptions", "name")

ISelectionModel = namedtuple("ISelectionModel", "last_selected_id selected_ids")

IVec2 = namedtuple("IVec2", "x y")
ICrop = namedtuple("ICrop", "bottom left right top")



from ..apibase import SlobsService, Event
from .typedefs import IStreamingState, _convert_time


class StreamingService(SlobsService):
    def __init__(self, connection):
        super().__init__(connection)

        self.recording_status_change = Event(
            connection, "recordingStatusChange", self.__class__.__name__
        )

        self.replay_buffer_status_change = Event(
            connection, "replayBufferStatusChange", self.__class__.__name__
        )
        self.streaming_status_change = Event(
            connection, "streamingStatusChange", self.__class__.__name__
        )

    async def get_model(self):
        response = await self._connection.command("getModel", self._prepared_params())
        return IStreamingState(
            recording_status=response["recordingStatus"],
            recording_status_time=_convert_time(response["recordingStatusTime"]),
            replay_buffer_status=response["replayBufferStatus"],
            replay_buffer_status_time=_convert_time(response["replayBufferStatusTime"]),
            streaming_status=response["streamingStatus"],
            streaming_status_time=_convert_time(response["streamingStatusTime"]),
        )

    async def save_replay(self):
        response = await self._connection.command("saveReplay", self._prepared_params())
        self._check_empty(response)

    async def start_replay_buffer(self):
        response = await self._connection.command(
            "startReplayBuffer", self._prepared_params()
        )
        self._check_empty(response)

    async def stop_replay_buffer(self):
        response = await self._connection.command(
            "stopReplayBuffer", self._prepared_params()
        )
        self._check_empty(response)

    async def toggle_recording(self):
        response = await self._connection.command(
            "toggleRecording", self._prepared_params()
        )
        self._check_empty(response)

    # Warning: According to the API, this may never return?!
    # Warning: If called too quickly, may fail without notice.
    async def toggle_streaming(self):
        response = await self._connection.command(
            "toggleStreaming", self._prepared_params()
        )
        self._check_empty(response)



from typing import Optional

from ..apibase import SlobsService
from .audiosource import AudioSource
from .factories import audiosource_factory


class AudioService(SlobsService):
    async def get_source(self, source_id) -> Optional[AudioSource]:
        response = await self._connection.command(
            "getSource", self._prepared_params([source_id])
        )
        if not response:
            return None
        return audiosource_factory(self._connection, response)

    async def get_sources(self) -> list[AudioSource]:
        response = await self._connection.command("getSources", self._prepared_params())
        return [
            audiosource_factory(self._connection, audio_source_dict)
            for audio_source_dict in response
        ]

    async def get_sources_for_current_scene(self) -> list[AudioSource]:
        response = await self._connection.command(
            "getSourcesForCurrentScene", self._prepared_params()
        )
        return [
            audiosource_factory(self._connection, audio_source_dict)
            for audio_source_dict in response
        ]

    async def get_sources_for_scene(self, scene_id) -> list[AudioSource]:
        response = await self._connection.command(
            "getSourcesForScene", self._prepared_params([scene_id])
        )
        return [
            audiosource_factory(self._connection, audio_source_dict)
            for audio_source_dict in response
        ]



from __future__ import annotations  # Postponed eval of annotations. Fixed in 3.10
from typing import Optional, Any, Dict

from .typedefs import TObsFormData, TSourceType, ISourceModel
from ..apibase import SlobsClass
from .factories import source_factory, register


class Source(SlobsClass):
    # There is an undocumented Configurable field. Sharing here without knowing what
    # it means.

    def __init__(
        self,
        connection,
        resource_id: str,
        source_id: str,
        async_: bool,
        audio: bool,
        channel: Optional[int],
        configurable: Optional[bool],
        do_not_duplicate: bool,
        height: float,
        id_: str,
        muted: bool,
        name_: str,
        type_: TSourceType,
        video: bool,
        width: float,
    ):
        super().__init__(connection, resource_id=resource_id, source_id=source_id)

        self._async = async_
        self._audio = audio
        self._channel = channel
        self._configurable = configurable
        self._do_not_duplicate = do_not_duplicate
        self._height = height
        self._id = id_
        self._muted = muted
        self._name = name_
        self._type = type_
        self._video = video
        self._width = width
        # Warning may be out of date if changed on the server.

    def __str__(self):
        return f"{self.__class__.__name__}({self._source_id}, {self._name!r})"

    @property
    def async_(self):
        return self._async

    @property
    def audio(self):
        return self._audio

    @property
    def channel(self):
        return self._channel

    @property
    def configurable(self):
        return self._configurable

    @property
    def do_not_duplicate(self):
        return self._do_not_duplicate

    @property
    def height(self):
        return self._height

    @property
    def id(self):
        return self._id

    @property
    def muted(self):
        return self._muted

    @property
    def name(self):
        return self._name

    @property
    def type_(self):
        return self._type

    @property
    def video(self):
        return self._video

    @property
    def width(self):
        return self._width

    async def duplicate(self) -> Source:
        response = await self._connection.command("duplicate", self._prepared_params())
        return source_factory(self._connection, response)

    async def get_model(self) -> ISourceModel:
        response = await self._connection.command("getModel", self._prepared_params())
        return ISourceModel(
            async_=response["async"],
            audio=response["audio"],
            channel=response.get("channel", None),
            do_not_duplicate=response["doNotDuplicate"],
            height=response["height"],
            id=response["id"],
            muted=response["muted"],
            name=response["name"],
            source_id=response["sourceId"],
            type_=response["type"],
            video=response["video"],
            width=response["width"],
        )

    async def get_properties_form_data(self) -> TObsFormData:
        response = await self._connection.command(
            "getPropertiesFormData", self._prepared_params()
        )
        return response

    async def get_settings(self) -> Dict[Any]:
        response = await self._connection.command(
            "getSettings", self._prepared_params()
        )
        return response

    async def has_props(self) -> bool:
        response = await self._connection.command("hasProps", self._prepared_params())
        return response

    async def refresh(self) -> None:
        response = await self._connection.command("refresh", self._prepared_params())
        self._check_empty(response)

    async def set_name(self, new_name) -> None:
        response = await self._connection.command(
            "setName", self._prepared_params([new_name])
        )
        self._check_empty(response)

    async def set_properties_form_data(self, properties: TObsFormData) -> None:
        response = await self._connection.command(
            "setPropertiesFormData", self._prepared_params([properties])
        )
        self._check_empty(response)

    async def update_settings(self, settings: Dict[Any]) -> None:
        response = await self._connection.command(
            "updateSettings", self._prepared_params([settings])
        )
        self._check_empty(response)


register(Source)



from ..apibase import SlobsService, Event
from .scene import Scene
from .factories import scenenode_factory, scene_factory
from .typedefs import ISceneModel


class ScenesService(SlobsService):
    def __init__(self, connection):
        super().__init__(connection)

        self.item_added = Event(connection, "itemAdded", self.__class__.__name__)
        self.item_removed = Event(connection, "itemRemoved", self.__class__.__name__)
        self.item_updated = Event(connection, "itemUpdated", self.__class__.__name__)
        self.scene_added = Event(connection, "sceneAdded", self.__class__.__name__)
        self.scene_removed = Event(connection, "sceneRemoved", self.__class__.__name__)
        self.scene_switched = Event(
            connection, "sceneSwitched", self.__class__.__name__
        )

    async def active_scene(self):
        response = await self._connection.command(
            "activeScene", self._prepared_params()
        )
        return scene_factory(self._connection, response)

    async def active_scene_id(self):
        response = await self._connection.command(
            "activeSceneId", self._prepared_params()
        )
        return response

    async def create_scene(self, scene_name: str) -> Scene:
        response = await self._connection.command(
            "createScene", self._prepared_params([scene_name])
        )
        return scene_factory(self._connection, response)

    async def get_scene(self, scene_id: str) -> Scene:
        response = await self._connection.command(
            "activeScene", self._prepared_params([scene_id])
        )
        return scene_factory(self._connection, response)

    async def get_scenes(self):
        response = await self._connection.command("getScenes", self._prepared_params())
        return [scene_factory(self._connection, subitem) for subitem in response]

    # Warning: In Studio Mode, this won't take immediate effect.
    # See TransitionsService to detect Studio Mode and to execute the transition.
    async def make_scene_active(self, scene_id: str) -> bool:
        response = await self._connection.command(
            "makeSceneActive", self._prepared_params([scene_id])
        )
        return response

    async def remove_scene(self, id_: str) -> ISceneModel:
        response = await self._connection.command(
            "removeScene", self._prepared_params([id_])
        )
        return scene_factory(self._connection, response)



import asyncio
from collections import defaultdict
import json
import logging
import time
from typing import Any, Optional

from websocket import create_connection, WebSocketTimeoutException

from .config import ConnectionConfig, config_from_ini
from .pubsubhub import PubSubHub, SubscriptionPreferences


class AuthenticationFailure(Exception):
    pass


class ProtocolError(Exception):
    pass


class _SlobsWebSocket:

    """
    _SlobsWebSocket is a class internal to the API. It is not used by
    a client.

    It is responsible for:
        Maintaining the websocket instance.
        Authentication.
        Sending and receiving low_level messages.
        Formatting a JSONRPC message.

    It does NOT use asyncio.
    """

    logger = logging.getLogger("slobsapi._SlobsWebSocket")

    def __init__(self, connection_config: ConnectionConfig, on_close=None):

        self.url = (
            f"ws://{connection_config.domain}:{connection_config.port}/api/websocket"
        )
        self.token = connection_config.token
        self._on_close = on_close
        try:
            self.socket = create_connection(self.url, timeout=20)
        except ConnectionRefusedError as e:
            raise ProtocolError("Couldn't connect. Is StreamLabs OBS running? %s" % e)
        self._authenticate()

    def is_alive(self) -> bool:
        return bool(self.socket)

    def send_message(self, id_, method, params) -> None:
        message_json = json.dumps(self._build_params_dict(id_, method, params))
        self.logger.debug("Sending:  %s", message_json)
        self.socket.send(message_json)

    def receive_message(self):
        while self.socket:
            try:
                raw_message = self.socket.recv()
                self.logger.debug("Received: %s", raw_message)
                if not raw_message:
                    break
                else:
                    try:
                        result = json.loads(raw_message)
                    except json.JSONDecodeError as json_error:
                        raise ProtocolError("%s from %s" % (json_error, raw_message))
                    except TypeError as type_error:
                        raise ProtocolError(type_error)
                    return result
            except WebSocketTimeoutException:
                self.logger.debug("Retrying after timeout.")
            except OSError as e:
                self.logger.warning(
                    "OSError received. Probably socket was closed. Retrying: %s", e
                )
                time.sleep(1)  # To prevent busy-loop
        self.close()
        return None

    def close(self) -> None:
        if self.socket:
            self.socket.close()
            self.socket = None
            if self._on_close:
                self._on_close()

    def _authenticate(self) -> None:
        message_id = "auth_request"
        self.send_message(
            id_=message_id,
            method="auth",
            params=dict(resource="TcpServerService", args=[self.token]),
        )
        response = self.receive_message()

        if response["id"] != message_id:
            raise ProtocolError("Response id mismatch: %s" % response)
        if "result" not in response or response["result"] is not True:
            raise AuthenticationFailure("%s" % response)

    @staticmethod
    def _build_params_dict(id_, method, params):
        request = {"jsonrpc": "2.0", "id": id_, "method": method, "params": params}
        return request


# noinspection PyBroadException
class SlobsConnection:
    """
    SlobsConnection is responsible for:
        Maintaining a SlobsWebsocket
        Assigning unique message ids to messages.
        Allowing users to subscribe to
            single replies
            ongoing resources
        Actively receiving responses
        Some sort of quit.
        Handling the transition from sync to async.
    """

    logger = logging.getLogger("slobsapi.SlobsConnection")

    def __init__(self, connection_config: Optional[ConnectionConfig] = None):

        connection_config = connection_config or config_from_ini()
        if not connection_config:
            raise ProtocolError("Connection not configured.")

        self.websocket = _SlobsWebSocket(connection_config, on_close=None)

        self._response_listeners = dict()
        self._event_listeners = defaultdict(lambda: set())
        self._undelivered_response = dict()
        self._undelivered_events = dict()
        self.hub = PubSubHub()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    def is_alive(self):
        return bool(self.websocket) and self.websocket.is_alive()

    async def _send(self, method, params):
        """
        Send message (in separate thread, so blocking doesn't hold up other
        coroutines) and wait for it to be sent.
        Returns the message_id used.
        """
        message_id = _message_id_factory.next()
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            lambda: self.websocket.send_message(message_id, method, params),
        )
        return message_id

    async def _receive_message(self):
        """
        Wait for received message (in separate thread).
        """
        if self.websocket:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, self.websocket.receive_message)

    async def _receive_and_dispatch(self):
        """
        Keep receiving messages and passing them on to the relevant queue
        for the consumers to pick up.

        This needs to be awaited to keep messages pumping.
        """
        try:
            while self.websocket and self.websocket.is_alive():
                try:
                    message = await self._receive_message()

                    if message:
                        if "id" in message and message["id"] is not None:
                            # This is a response to an explicit request.
                            key = message["id"]
                            await self.hub.publish(key=key, message=message)
                        else:
                            # This is a response to an subscription.
                            key = message["result"]["resourceId"]
                            data = message["result"].get("data", None)

                            await self.hub.publish(key=key, message=data)
                except AttributeError:
                    if not self.hub:
                        # Connection was closed mid-receive.
                        break
                    raise
        except Exception:
            self.logger.exception("_receive_and_dispatch failing.")
            raise

        await self.close()

    # Give a nicer name for clients that don't need the details.
    background_processing = _receive_and_dispatch

    async def command(self, method, params):
        """
        Send a command that expects a response.
        Wait for result. If the result returns a promise, wait for the promise.
        """
        self.logger.debug("Command being sent: %s(%s)", method, params)
        message_id = await self._send(method, params)
        final_response_queue = asyncio.Queue()  # Callback will push the response here.

        async def callback(key, value):
            try:
                await self._command_accept_result_of_promise(
                    key, value, final_response_queue
                )
            except Exception:
                self.logger.exception("command.callback has failed!")

        assert asyncio.iscoroutinefunction(callback)
        await self.hub.subscribe(key=message_id, callback_coroutine=callback)
        response = await final_response_queue.get()
        await self.hub.unsubscribe(key=message_id, callback_coroutine=callback)

        if isinstance(response, Exception):
            raise response
        else:
            return response

    async def _command_accept_result_of_promise(self, _, response, queue):
        try:
            if "error" in response:
                await queue.put(ProtocolError(response["error"]))
                return

            if "result" not in response:
                await queue.put(ProtocolError("No result found: %s" % response))
                return

            result = response["result"] or {}

            if isinstance(result, dict) and result.get("emitter", "NA") == "PROMISE":
                # Server is informing that the result may be some time.
                # It will come as an event.
                key = result["resourceId"]
                response_as_event_queue = asyncio.Queue()

                async def callback(k, v):
                    try:
                        await self._command_accept_event_as_response(
                            k, v, response_as_event_queue
                        )
                    except Exception:
                        self.logger.exception(
                            "_command_accept_result_of_promise.callback has failed!"
                        )

                await self.hub.subscribe(key, callback)
                response = await response_as_event_queue.get()
                await self.hub.unsubscribe(key, callback)
                await queue.put(response)
            else:
                await queue.put(result)
        except Exception:
            self.logger.exception("_command_accept_result_of_promise has failed!")

    @staticmethod
    async def _command_accept_event_as_response(_, response, queue) -> None:
        try:
            await queue.put(response)
        except Exception:
            self.logger.exception("_command_accept_event_as_response has failed!")

    async def subscribe(
        self,
        method,
        params,
        callback_coroutine,
        subscription_preferences=SubscriptionPreferences(),
    ) -> Any:
        """
        Send a command that subscribes to events.
        Wait for result. The result should include resource_id.
        Return the resource_id, and continue to have callback_coroutine called as the
        events trigger.
        """

        # Command may be unnecessary if there is already a subscription, but we don't
        # know the resource_id here until we ask.
        response = await self.command(method, params)
        if not "_type" in response or response["_type"] != "SUBSCRIPTION":
            raise ProtocolError("Badly formed subscription response: %s" % response)
        try:
            resource_id = response["resourceId"]
        except KeyError:
            raise ProtocolError("Badly formed subscription response: %s" % response)

        await self.hub.subscribe(
            key=resource_id,
            callback_coroutine=callback_coroutine,
            subscription_preferences=subscription_preferences,
        )
        return resource_id

    async def unsubscribe(self, resource_id, callback_coroutine) -> None:
        """
        Send a command that unsubscribes to events (resource should be provided
        in params dict).
        Also, unhook callback_coroutine from hub.
        """

        await self.hub.unsubscribe(
            key=resource_id, callback_coroutine=callback_coroutine
        )
        # Only tell server if no-one else is interested.
        if not self.hub.has_subscribers(resource_id):
            response = await self.command("unsubscribe", {"resource": resource_id})
            if not response:
                raise ProtocolError("Unsubscribe failed.")

    async def close(self):
        self.logger.debug("Request to close connection.")
        if self.websocket:
            socket_to_close = self.websocket
            self.websocket = None  # So we don't recurse.
            socket_to_close.close()

        if self.hub:
            await self.hub.close()
            self.hub = None


class _MessageIdFactory:
    def __init__(self):
        self._message_id = 0

    def next(self):
        self._message_id += 1
        return self._message_id


_message_id_factory = _MessageIdFactory()



from .config import ConnectionConfig, config_from_ini_else_stdin, config_from_ini
from .connection import AuthenticationFailure, ProtocolError, SlobsConnection
from .pubsubhub import SubscriptionPreferences, CLOSED, UNSUBSCRIBED
from .slobs.audioservice import AudioService
from .slobs.notificationsservice import NotificationsService
from .slobs.performanceservice import PerformanceService
from .slobs.scenecollectionservice import SceneCollectionsService
from .slobs.scenesservice import ScenesService
from .slobs.selectionservice import SelectionService
from .slobs.sourcesservice import SourcesService
from .slobs.streamingservice import StreamingService
from .slobs.transitionsservice import TransitionsService
from .slobs.typedefs import (
    NotificationSubType,
    NotificationType,
    MonitoringType,
    TSceneNodeType,
    ISourceAddOptions,
    ISceneCollectionCreateOptions,
    ITransform,
    IVec2
)

__all__ = [
    "AuthenticationFailure",
    "ProtocolError",
    "SlobsConnection",
    "SubscriptionPreferences",
    "AudioService",
    "NotificationsService",
    "PerformanceService",
    "SceneCollectionsService",
    "SelectionService",
    "SourcesService",
    "StreamingService",
    "TransitionsService",
    "NotificationSubType",
    "NotificationType",
    "TSceneNodeType",
    "ISourceAddOptions",
    "ISceneCollectionCreateOptions",
    "ITransform",
]



"""
    A Publish-Subscribe service.

        - Messages are associated with a key. Only subscribers to that key will receive
          the message.
            - No support for wildcards.
        - Optionally informs subscriber that the queue is being shutdown.
        - Optionally informs subscriber that they have been unsubscribed.
        - Maintain a (limited) buffer of undelivered messages in case a message
          is received before subscriber can subscribe.
        
        Only one asyncio mechanism is supported. Each subscriber must provide a 
        callback_coroutine co-routine.
            - Will be executed in the context of the publisher's loop.
            - Must return fast to prevent stalling the publisher.
            - Callback takes two parameters: key and message.
            - Unsubscribe notifications will have UNSUBSCRIBED as the message.
            - Close notifications will have CLOSED as the key and message.
"""
from __future__ import annotations
import asyncio
from dataclasses import dataclass
import logging
from typing import Any, Callable, Coroutine

LOGGER = logging.getLogger("slobsapi.pubsubhub")


@dataclass(frozen=True)
class SubscriptionPreferences:
    notify_on_unsubscribe: bool = False
    notify_on_close: bool = False


UNSUBSCRIBED = object()
CLOSED = object()


class PubSubHub:

    Callback = Callable[[Any, Any], Coroutine]  # The return value is a coroutine.

    def __init__(self):

        self._subscribers_by_key: dict[
            str, dict[PubSubHub.Callback, SubscriptionPreferences]
        ] = dict()
        # Map from key ->
        #     (dict mapping from callback_coroutine -> SubscriptionPreferences)

        self._undelivered_messages: dict[str, list[any]] = dict()
        # Map from key -> list of messages

    async def subscribe(
        self,
        key: Any,
        callback_coroutine: PubSubHub.Callback,
        subscription_preferences: SubscriptionPreferences = SubscriptionPreferences(),
    ):
        assert asyncio.iscoroutinefunction(callback_coroutine), "Callback must be async"
        if key not in self._subscribers_by_key:
            self._subscribers_by_key[key] = {
                callback_coroutine: subscription_preferences
            }
        else:
            self._subscribers_by_key[key][callback_coroutine] = subscription_preferences

        # If we have any undelivered messages to this key, send them now.

        # Check if response already arrived.
        if key in self._undelivered_messages:
            # Response already arrived!
            LOGGER.debug(
                "Backlog of messages being cleared: %s", self._undelivered_messages[key]
            )
            for message in self._undelivered_messages[key]:
                asyncio.ensure_future(callback_coroutine(key, message))

            del self._undelivered_messages[key]

    async def unsubscribe(self, key: Any, callback_coroutine: Callback):
        if key in self._subscribers_by_key:
            if callback_coroutine in self._subscribers_by_key[key]:
                prefs = self._subscribers_by_key[key][callback_coroutine]
                del self._subscribers_by_key[key][callback_coroutine]
                if not self._subscribers_by_key[key]:
                    del self._subscribers_by_key[key]
                if prefs.notify_on_unsubscribe:
                    asyncio.ensure_future(callback_coroutine(key, UNSUBSCRIBED))

    def has_subscribers(self, key):
        return key in self._subscribers_by_key

    async def publish(self, key: Any, message):
        subscribers = self._subscribers_by_key.get(key, {})

        if not subscribers:
            # Add to undelivered list.
            if key not in self._undelivered_messages:
                self._undelivered_messages[key] = [message]
            elif len(self._undelivered_messages[key]) >= 10:
                LOGGER.warning(
                    "Discarding undelivered message to %s, due to backlog. "
                    "Likely server subscriptions don't match client "
                    "subscriptions.",
                    key,
                )
            else:
                self._undelivered_messages[key].append(message)
        else:
            coroutines = subscribers.keys()
            assert all(
                asyncio.iscoroutinefunction(coroutine) for coroutine in coroutines
            )
            for coroutine in coroutines:
                asyncio.ensure_future(coroutine(key, message))

    async def close(self):
        subscribers_to_notify = [
            callback
            for key in self._subscribers_by_key
            for (callback, prefs) in self._subscribers_by_key[key].items()
            if prefs.notify_on_close
        ]
        for coroutine in subscribers_to_notify:
            asyncio.ensure_future(coroutine(CLOSED, CLOSED))



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PySLOBS",
    version="1.2.3",
    author="Julian-O",
    # author_email="",
    description="Python wrapper to StreamLabs OBS API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Julian-O/pyslobs",
    packages=setuptools.find_packages(),
    tests_require=[],
    install_requires=["websocket_client"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Topic :: Multimedia :: Video",
    ],
    python_requires=">=3.9",
)



from typing import Optional
from ..apibase import SlobsService, Event
from .typedefs import (
    ISceneCollectionsManifestEntry,
    ISceneCollectionSchema,
    ISceneCollectionCreateOptions,
)
from .factories import source_factory, sceneitem_factory


def iscenecollectionmanifestentry_factory(json_dict):
    return ISceneCollectionsManifestEntry(id=json_dict["id"], name=json_dict["name"])


def iscenecollectionschema_factory(connection, json_dict):
    return ISceneCollectionSchema(
        id=json_dict["id"],
        name=json_dict["name"],
        # The field is called scenes, but doesn't actually contain scenes.
        # Appears to be a novel type that references scenes. Returning the raw dict.
        scenes=json_dict["scenes"],
        # The field is called sources, but doesn't actually contain sources.
        # Appears to be a novel type that references sources. Returning the raw dict.
        sources=[json_dict["sources"]],
    )


class SceneCollectionsService(SlobsService):
    def __init__(self, connection):
        super().__init__(connection)

        self.collection_added = Event(
            connection, "collectionAdded", self.__class__.__name__
        )
        self.collection_removed = Event(
            connection, "collectionRemoved", self.__class__.__name__
        )
        self.collection_switched = Event(
            connection, "collectionSwitched", self.__class__.__name__
        )
        self.collection_updated = Event(
            connection, "collectionUpdated", self.__class__.__name__
        )
        self.collection_will_switch = Event(
            connection, "collectionWillSwitch", self.__class__.__name__
        )

    async def active_collection(self):
        response = await self._connection.command(
            "activeCollection", self._prepared_params([])
        )
        return iscenecollectionmanifestentry_factory(response)

    async def collections(self):
        response = await self._connection.command(
            "collections", self._prepared_params([])
        )
        return [iscenecollectionmanifestentry_factory(subitem) for subitem in response]

    async def create(self, options: ISceneCollectionCreateOptions):
        response = await self._connection.command(
            "create", self._prepared_params([{"name": options.name}])
        )
        return iscenecollectionmanifestentry_factory(response)

    async def delete(self, id: Optional[str] = None) -> None:
        response = await self._connection.command("delete", self._prepared_params([id]))
        self._check_empty(response)

    async def fetch_scene_collections_schema(self):
        # Note: Intermittently raises ProtocolError (code=-32600 INVALID_REQUEST)
        # possibly if called too quickly after other operations?
        response = await self._connection.command(
            "fetchSceneCollectionsSchema", self._prepared_params([])
        )
        return [
            iscenecollectionschema_factory(self._connection, subitem)
            for subitem in response
        ]

    async def load(self, id: str) -> None:
        response = await self._connection.command("load", self._prepared_params([id]))
        self._check_empty(response)

    async def rename(self, new_name: str, id: str) -> None:
        response = await self._connection.command(
            "rename", self._prepared_params([new_name, id])
        )
        self._check_empty(response)



from ..apibase import SlobsClass
from .typedefs import IFader, IAudioSourceModel, MonitoringType
from .factories import register


class AudioSource(SlobsClass):
    async def get_model(self):
        response = await self._connection.command("getModel", self._prepared_params())
        return IAudioSourceModel(
            audio_mixers=response["audioMixers"],
            fader=IFader(
                db=response["fader"]["db"],
                deflection=response["fader"]["deflection"],
                mul=response["fader"]["mul"],
            ),
            force_mono=response["forceMono"],
            mixer_hidden=response["mixerHidden"],
            monitoring_type=MonitoringType(response["monitoringType"]),
            muted=response["muted"],
            name=response["name"],
            source_id=response["sourceId"],
            sync_offset=response["syncOffset"],
        )

    async def set_deflection(self, deflection):
        # I don't know what legal values are (small integers?), but illegal values
        # are simply ignored.
        response = await self._connection.command(
            "setDeflection", self._prepared_params([deflection])
        )
        self._check_empty(response)

    async def set_muted(self, muted):
        response = await self._connection.command(
            "setMuted", self._prepared_params([muted])
        )
        self._check_empty(response)


# Dependency Injection
register(AudioSource)



from typing import Optional, Any
from ..apibase import SlobsClass
from .scenenode import (
    SceneNode,
)
from .source import Source
from .factories import (
    scenenode_factory,
    sceneitem_factory,
    sceneitemfolder_factory,
    selection_factory,
    source_factory,
    register,
)
from .typedefs import ISceneNodeAddOptions, ISourceAddOptions, TSourceType


class Scene(SlobsClass):
    def __init__(
        self,
        connection,
        resource_id: str,
        source_id: str,
        id: str,
        name: str,
        nodes: list[SceneNode],
    ):
        super().__init__(connection, resource_id=resource_id, source_id=source_id)

        self._name = name
        # Warning may be out of date if changed on the server.
        # Use set_name to change

        self._id = id
        self._nodes = nodes

    def __str__(self):
        return f"{self.__class__.__name__}({self._resource_id}, {self.name!r})"

    @property
    def name(self):
        return self._name

    @property
    def nodes(self):
        return self._nodes

    @property
    def id(self):
        return self._id

    async def add_file(self, path, folder_id: Optional[str] = None) -> SceneNode:
        response = await self._connection.command(
            "addFile", self._prepared_params([str(path), folder_id])
        )
        raise scenenode_factory(self._connection, response)

    async def add_source(
        self, source_id: str, options: Optional[ISceneNodeAddOptions] = None
    ) -> SceneNode:
        options_dict = {}
        if options:
            if options.id is not None:
                option_dict["id"] = options.id
            if options.source_add_options is not None:
                source_add_options = {}
                if options.source_add_options.channel is not None:
                    source_add_options["channel"] = options.source_add_options.channel
                if options.source_add_options.is_temporary is not None:
                    source_add_options[
                        "isTemporary"
                    ] = options.source_add_options.is_temporary
                options_dict["sourceAddOptions"] = source_add_options
        response = await self._connection.command(
            "addSource", self._prepared_params([source_id, options_dict])
        )
        return sceneitem_factory(self._connection, response)

    async def can_add_source(self, source_id: str) -> bool:
        response = await self._connection.command(
            "canAddSource", self._prepared_params([source_id])
        )
        return response

    async def clear(self):
        response = await self._connection.command("clear", self._prepared_params())
        self._check_empty(response)

    async def create_and_add_source(
        self, name: str, type_: TSourceType, settings: Optional[dict[Any]] = None
    ):  # -> SceneItem:
        response = await self._connection.command(
            "createAndAddSource", self._prepared_params([name, type_, settings])
        )
        return sceneitem_factory(self._connection, response)

    async def create_folder(self, name: str):  # -> SceneItem:
        response = await self._connection.command(
            "createFolder", self._prepared_params([name])
        )
        return sceneitemfolder_factory(self._connection, response)

    async def get_folder(self, scene_folder_id: str):  # -> SceneItemFolder:
        response = await self._connection.command(
            "getFolder", self._prepared_params([scene_folder_id])
        )
        return sceneitemfolder_factory(self._connection, response)

    async def get_folders(self):  # -> list[SceneItemFolder]:
        response = await self._connection.command("getFolders", self._prepared_params())
        return [
            sceneitemfolder_factory(self._connection, subitem) for subitem in response
        ]

    async def get_item(self, scene_folder_id: str):  # -> SceneItem:
        response = await self._connection.command(
            "getItem", self._prepared_params([scene_folder_id])
        )
        return sceneitem_factory(self._connection, response)

    async def get_items(self):  # -> list[SceneItem]:
        response = await self._connection.command("getItems", self._prepared_params())
        return [sceneitem_factory(self._connection, subitem) for subitem in response]

    async def get_model(self):
        response = await self._connection.command("getModel", self._prepared_params())
        return scene_factory(self._connection, response)

    async def get_nested_items(self):  # -> Source
        response = await self._connection.command(
            "getNestedItems", self._prepared_params()
        )
        return [sceneitem_factory(self._connection, source) for source in response]

    async def get_nested_scenes(self):  # -> Source
        response = await self._connection.command(
            "getNestedScenes", self._prepared_params()
        )
        return [scene_factory(self._connection, source) for source in response]

    async def get_nested_sources(self):  # -> Source
        response = await self._connection.command(
            "getNestedSources", self._prepared_params()
        )
        return [source_factory(self._connection, source) for source in response]

    async def get_node(self, scene_node_id):  # -> Source
        response = await self._connection.command(
            "getNode", self._prepared_params([scene_node_id])
        )
        return scenenode_factory(self._connection, source)

    async def get_node_by_name(self, name: str) -> Optional[Source]:
        response = await self._connection.command(
            "getNodeByName", self._prepared_params([name])
        )
        if not response:
            return None
        else:
            return scenenode_factory(self._connection, response)

    async def get_nodes(self):  # -> Source
        response = await self._connection.command("getNodes", self._prepared_params())
        result = [scenenode_factory(self._connection, source) for source in response]
        # Update attributes as a side-effect
        self._nodes = result
        return result

    async def get_root_nodes(self):
        response = await self._connection.command(
            "getRootNodes", self._prepared_params()
        )
        return [scenenode_factory(self._connection, node) for node in response]

    async def get_selection(self, ids: Optional[list[str]] = None):
        response = await self._connection.command(
            "getSelection", self._prepared_params(ids)
        )
        return selection_factory(self._connection, response)

    async def get_source(self):
        response = await self._connection.command("getSource", self._prepared_params())
        return source_factory(self._connection, response)

    async def make_active(self):
        response = await self._connection.command("makeActive", self._prepared_params())
        self._check_empty(response)

    async def remove(self):
        response = await self._connection.command("remove", self._prepared_params())
        self._check_empty(response)

    async def remove_folder(self, folder_id):
        response = await self._connection.command(
            "removeFolder", self._prepared_params([folder_id])
        )
        self._check_empty(response)

    async def remove_item(self, scene_item_id):
        response = await self._connection.command(
            "removeItem", self._prepared_params([scene_item_id])
        )
        self._check_empty(response)

    async def set_name(self, new_name):
        response = await self._connection.command(
            "setName", self._prepared_params([new_name])
        )
        self._check_empty(response)

        # Update local cache.
        self._name = new_name


register(Scene)
